from http.server import BaseHTTPRequestHandler

from src.app.Router import Router

class Server(BaseHTTPRequestHandler):
  router = Router()

  def log_request(self, format, *args):
      return
  def log_message(self, format, *args):
    return

  def do_HEAD(self):
    return
    
  def do_GET(self):
    try:
      self.respond()
    except:
      self.send_response(400)
    
  def do_POST(self):
    return
    
  def handle_http(self):
    if self.path in self.router.routes:
      response = self.router.routes[self.path]()
    else:
      response = self.router.static(self.path)

    self.send_response(response.status)
    self.send_header('Content-type', response.type)
    self.end_headers()

    try:
      return bytes(response.data, 'utf-8')
    except:
      return response.data
    
  def respond(self):
    content = self.handle_http()
    self.wfile.write(content)


import subprocess
import json
from threading import Thread

from src.app.path import get_path
from src.app.utils import check_npm_availability

thread = None
updates_data = {}

def pull_updates():
  print('getting updates data...')
  global updates_data
  try:
    path = get_path()
    command = ['npm', 'outdated', '--json', '--all']
    result = subprocess.run(command, capture_output=True, cwd=path).stdout
    updates_data = json.loads(result)
    print('updates data received successfully', flush=True)
  except:
    pass

def get_updates():
  global updates_data
  global thread
  if not updates_data:
    if check_npm_availability() != 200:
      return
    if not thread:
      thread = Thread(target=pull_updates, daemon=True)
      thread.start()
    thread.join()
  return updates_data


import copy

from src.app.utils import to_dict

attributes = ['name', 'devDependencies' 'dependencies', 'author', 'description', 'keywords', 'license', 'repository', 'version', 'path']

class Lib:
  def __init__(self, data, path):
    self.errors = []
    self.connections = []
    self.unprocessed_deps = {}
    self.peerDependencies = {}
    self.path = path

    for k, v in data.items():
      if(k in attributes):
        setattr(self, k, v)
      # dev dependencies should be added only for root package because only those are actually installed
      # path == '' means root package (empty node_modules path)
      if k == 'dependencies' or (k == 'devDependencies' and path == ''):
        self.unprocessed_deps.update(copy.deepcopy(v))

  def process_dependency(self):
    if len(self.unprocessed_deps.keys()):
      deps = list(self.unprocessed_deps.keys())
      child = deps[0]
      version = self.unprocessed_deps[child]
      self.unprocessed_deps.pop(child, None)
      return (child, version)

  def add_error(self, error):
    exist = next((item for item in self.errors if item['type'] == error['type'] and item['lib'] == error['lib']), None)

    if not exist:
      self.errors.append(error)
  
  def add_connection(self, data):
    exist = next((item for item in self.connections if item['name'] == data['name']), None)
    if exist:
      return
    self.connections.append(data)

  def get_data(self):
    result = {}
    for k in attributes:
      if self[k]:
        result[k] = self[k]
    result['connections'] = self.connections
    result['errors'] = to_dict(self.errors)
    return result

  def __getitem__(self, item):
    return getattr(self, item, None)


import os
import json
from mimetypes import guess_type

from src.app.classes.Response import Response
from src.app.utils import open_file

from src.app.data.dependencies import get_dependencies, get_issues
from src.app.data.updates import get_updates
from src.app.data.vulnerabilities import get_vulnerabilities
from src.app.data.package import get_package_data
from src.app.data.deprecated import get_deprecated

class Router:
  def __init__(self):
    self.routes = {
      '/': self.root,
      '/api/dependencies': self.send_dependencies,
      '/api/vulnerabilities': self.send_vulnerabilities,
      '/api/updates': self.send_updates,
      '/api/issues': self.send_issues,
      '/api/package': self.send_package_data,
      '/api/deprecated': self.send_deprecated_data
    }

  def root(self):
    return self.static('/index.html')

  def send_dependencies(self):
    return Response('application/json', json.dumps(get_dependencies()))

  def send_deprecated_data(self):
    result = get_deprecated()
    if not result:
      return Response('application/json', 'network error', status = 502)
    return Response('application/json', json.dumps(result))

  def send_vulnerabilities(self):
    result = get_vulnerabilities()
    if not result:
      return Response('application/json', 'network error', status = 502)
    return  Response('application/json', json.dumps(result))
  
  def send_issues(self):
    return Response('application/json', json.dumps(get_issues()))

  def send_package_data(self):
    return Response('application/json', json.dumps(get_package_data()))

  def send_updates(self):
    result = get_updates()
    if not result:
      return Response('application/json', 'network error', status = 502)
    return Response('application/json', json.dumps(result))

  def static(self, path):
    static_folder = '/client'
    dir_path = os.path.dirname(os.path.realpath(__file__))
    final_path = dir_path + '/..' + static_folder + path
    if '..' in path:
      return Response('text/plain', '403 forbidden', status = 403)  

    isfile = os.path.isfile(final_path)

    if isfile:
      data = open_file(final_path)
      type = guess_type(final_path)
      return Response(type[0], data)
    else:
      return self.not_found()

  def not_found(self):
    return Response('text/plain', '404 not found', status = 404)   



from src.app.utils import open_json_file

from src.app.path import get_package_json_path

def get_package_data():
  package_json_path = get_package_json_path()
  return open_json_file(package_json_path)


import os

from src.app.utils import open_json_file, to_dict
from src.app.classes.Lib import Lib
from src.app.classes.Error import Error
from src.app.path import get_package_json_path, get_node_modules_path, get_path

# key: name, value: arr of package versions (Lib)
dependencies = {}
dev_deps = {}
root = None
max_depth = 0
loaded = False

def process_dependencies():
  print('parsing dependencies tree...', flush=True)

  global dependencies
  global root
  global max_depth

  package_json_path = get_package_json_path()
  node_modules_path = get_node_modules_path()

  root_data = open_json_file(package_json_path)
  root = Lib(root_data, '')
  # updatable
  stack = [root]

  # dependencies tree processing inmplemented with iterative way
  while len(stack):
    depth = len(stack)
  
    # checking and increasing max depth
    if depth > max_depth:
      max_depth = depth

    # getting a package for processing
    current = stack[depth - 1]

    # looking for circular dependency by checking is the same lib already in stack
    circular_dependency_index = find_circular_dependency(stack, current)

    # if circular dependency found put error in the parend and skip parsing this dependency
    if circular_dependency_index is not False:
      parent = stack[circular_dependency_index]
      parent.add_error(Error("circular", current.name, current.version))
      stack.pop()
      continue

    # adding package to packages list
    add_to_dependencies_list(current, dependencies)

    # dependencies parsing
    child = current.process_dependency()
    
    # TODO refactor
    # add check is version exists in array before searching in directories
    # if not processed dependency exist, getting dependency data
    if child:
      name, version = child
      child_folder_path = name
      
      # if another version already found check local package node_modules
      if is_local_dependency(current.path, name):
        child_folder_path = current.path + '/node_modules/' + name

      # path to dependency package.json file 
      child_path = node_modules_path + '/' + child_folder_path +  '/package.json'
      try:
        # getting data from child package.json file
        child_data = open_json_file(child_path)
        # appending child to the stack so its dependencies can be processed next
        stack.append(Lib(child_data, child_folder_path))
        # adding child to current lib connections
        current.add_connection({'name': child_data['name'], 'version': child_data['version']})
      except:
        # if package.json was not found adding error to the current package
        current.add_error(Error('missing', name, version))
    else:
      # removing current package from the stack if it has no unpocessed deps
      stack.pop(len(stack) - 1)
  print('dependencies tree parsed successfully', flush=True)

def is_version_exists(versions, version):
  return next((item for item in versions if item.version == version), False)

def is_local_dependency(current, name):
  node_modules_path = get_node_modules_path()
  path = node_modules_path + '/' + current + '/node_modules/' + name
  return os.path.isdir(path)

def find_circular_dependency(stack, current):
  for i, prev in enumerate(stack[:len(stack) - 1]):
    if current.name == prev.name:
      return i
  return False

def add_to_dependencies_list(current, dependencies):
  if current.name not in dependencies:
    dependencies[current.name] = [current]
  elif not is_version_exists(dependencies[current.name], current.version):
    dependencies[current.name].append(current)

# getting dependencies tree data
def get_dependencies():
  result = {}
  for key in dependencies:
    deps = []
    for l in dependencies[key]:
      deps.append(l.get_data())
    result[key] = deps
  return {
    'root': root.name,
    'dependencies': result,
    'depth': max_depth
  }

def get_issues():
  global dependencies
  result = {}
  for key in dependencies:
    for package in dependencies[key]:
      errors = package.errors
      for error in errors:
        if not error.type in result:
          result[error.type] = {}
        result[error.type][error.name] = to_dict(error)
  return result


from http.server import HTTPServer
from socketserver import ThreadingMixIn
import sys

from src.app.utils import port_check
from src.app.Server import Server
from src.app.data.dependencies import process_dependencies

def init():
  HOST_NAME = '127.0.0.1'
  PORT_NUMBER = 8080
  if len(sys.argv) == 3:
    PORT_NUMBER = int(sys.argv[2])

  isOpen = port_check(HOST_NAME, PORT_NUMBER)
  if not isOpen:
    print('ERROR: port ' + str(PORT_NUMBER) + ' is not available. Provide available port as the second command param', flush=True)
    exit()

  class ThreadingSimpleServer(ThreadingMixIn, HTTPServer):
    pass

  process_dependencies()
  httpd = ThreadingSimpleServer((HOST_NAME, PORT_NUMBER), Server)
  try:
    print('server is running on http://%s:%s' % (HOST_NAME, PORT_NUMBER), flush=True)
    httpd.serve_forever()
  except KeyboardInterrupt:
    print('\nExiting app...', flush=True)

  httpd.server_close()


# types circular/missing/peer
class Error:
  def __init__(self, type, lib, version = '', data = {}):
    self.name = lib
    self.type = type
    self.data = data
    self.version = version

  def __getitem__(self, item):
    return getattr(self, item, None)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['src', 'src.app', 'src.app.classes', 'src.app.data']

package_data = \
{'': ['*'],
 'src': ['client/*',
         'client/css/*',
         'client/font/*',
         'client/js/*',
         'client/js/components/*',
         'client/js/components/elements/*',
         'client/js/components/items/*',
         'client/js/components/popups/*',
         'client/js/lib/*',
         'client/js/plots/*',
         'client/js/workers/*',
         'client/static/*']}

install_requires = \
['requests>=2.28.2,<3.0.0']

entry_points = \
{'console_scripts': ['npm-cockpit = src.app.app:init']}

setup_kwargs = {
    'name': 'npm-cockpit',
    'version': '0.3.0',
    'description': 'Installable command-line tool with web interface which helps to track nodejs project dependencies state',
    'long_description': '# NPM-COCKPIT\n\nA user-friendly application for JavaScript developers to visualize the dependency tree of NPM packages and NodeJS applications. \n\nAllows to get statistical info about application dependencies tree state. Provides an interface for filtering and viewing information through convenient tables and diagrams.\n\n## Features\n\nVisual representation of the entire dependency tree of a project. With ability to look all the paths for a specific package. Dependency tree can be visualized as a tree or directed network chart\n\n![tree chart](https://chartexample.com/images/npm-cockpit/network.jpg)\n\nDetailed information about each package, including version, description, and related links.\n\n![packages list](https://chartexample.com/images/npm-cockpit/list.jpg)\n\nIdentify potential issues such as outdated, deprecated or vulnerable packages.\n\n![packages list](https://chartexample.com/images/npm-cockpit/deprecated.jpg)\n\n**AND MUCH MORE!**\n\n\n## Requirements\n\n### General\n- Terminal or command prompt access\n- Target application folder should contain node_modules folder with installed dependencies and package.json\n- Node.js and NPM installed\n\n### As NPM package\n- `python` command should be available\n\n## Usage\n\n### Command params\n\n`path` - a path to project folder with package.json and node_modules inside\n\n`port` - available local port to serve the app (default `8080`)\n\n### PIP\n`pip install npm-cockpit`\n\n`npm-cockpit [path] [port]`\n\n### NPM globally installed\n`npm install --global npm-cockpit`\n\n`npm-cockpit [path] [port]`\n\n### NPX\n`npx npm-cockpit [path] [port]`\n\n### NPM dependency in package\n`npm install npm-cockpit` and add the run script in the package json with proper params\n\n## Development\n- `git clone https://github.com/b0000ring/npm-cockpit.git`\n- `cd npm-cockpit`\n- `node index.js [path] [port]` or `python __main__.py [path] [port]`\n',
    'author': 'Alex Chirkin',
    'author_email': 'hello@alexchirkin.me',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



import sys
import os

try:
  sys.argv[1]
except:
  print('ERROR: Please, provide the path to folder with package.json and node_modules', flush=True)
  exit()

path = sys.argv[1]
package_json_path = path + '/package.json'
package_lock_path = path + '/package-lock.json'
node_modules_path = path + '/node_modules'

isNodeModulesFound = os.path.isdir(node_modules_path)
isPackageJSONFound = os.path.isfile(package_json_path)
isPackageLockFound = os.path.isfile(package_lock_path)

# provided path validation
if not(isPackageJSONFound):
  print('ERROR: Target folder should contain package.json file', flush=True)
  exit()
if not(isNodeModulesFound):
  print('ERROR: Target folder should contain node_modules folder', flush=True)
  exit()
if not(isPackageLockFound):
  print('WARNING: package-lock.json not found, cannot get vulnerabilities info', flush=True)

def get_path():
  global path
  return path

def get_package_json_path():
  global package_json_path
  return package_json_path

def get_node_modules_path():
  global node_modules_path
  return node_modules_path


import json
import socket
import urllib.request

def open_json_file(path):
  try:
    with open(path) as file:
      return json.load(file)
  except:
    pass
    # print(path)

def to_dict(obj):
  return json.loads(json.dumps(obj, default=lambda o: o.__dict__))

def open_file(path):
  try:
    with open(path) as file:
      return file.read()
  except:
     with open(path, 'rb') as file:
      return file.read()

def check_npm_availability():
  try:
    resp = urllib.request.urlopen("https://www.npm.com", timeout=10).getcode()
  except:
    resp = 502
  return resp

# TODO implement
def check_semver():
  None

def port_check(host, port):
  s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
  s.settimeout(1) #Timeout in case of port not open
  try:
    s.bind((host, port)) 
    return True
  except:
    return False



import concurrent.futures
import requests
from threading import Thread

from src.app.data.dependencies import dependencies
from src.app.utils import check_npm_availability

thread = None
deprecated_data = {}

def pull_package_data(name):
  global deprecated_data
  response = requests.get(url='https://registry.npmjs.org/' + name + '/latest', timeout=10)
  if response.status_code == 200:
    data = response.json()
    if(data['deprecated']):
      deprecated_data[name] = data['deprecated']

def get_deprecated_data():
  print('getting deprecation data...', flush=True)
  with concurrent.futures.ThreadPoolExecutor() as executor:
    futures = []
    for lib in dependencies: 
      futures.append(executor.submit(pull_package_data, name=lib))
    concurrent.futures.wait(futures)
  print('deprecation data received successfully', flush=True)

def get_deprecated():
  global deprecated_data
  global thread
  if not deprecated_data:
    if check_npm_availability() != 200:
      return
    if not thread:
      thread = Thread(target=get_deprecated_data)
      thread.start()
    thread.join()
  return deprecated_data


import subprocess
import json
from threading import Thread

from src.app.utils import check_npm_availability
from src.app.path import get_path

vulnerabilities_data = {}
thread = None

def pull_vulnerabilities():
  print('getting vulnerabilities data...', flush=True)
  global vulnerabilities_data
  try:
    path = get_path()
    command = ['npm', 'audit', '--json']
    result = subprocess.run(command, capture_output=True, cwd=path).stdout
    vulnerabilities_data = json.loads(result)
    print('vulnerabilities data received successfully', flush=True)
  except:
    pass


def get_vulnerabilities():
  global vulnerabilities_data
  global thread
  if not vulnerabilities_data:
    if check_npm_availability() != 200:
      return
    if not thread:
      thread = Thread(target=pull_vulnerabilities, daemon=True)
      thread.start()
    thread.join()
  return vulnerabilities_data
  
